from taf.utils import normalize_line_endings


def test_normalize_line_ending_extra_lines():
    test_content = b"""This is some text followed by two new lines


"""
    expected_content = b"This is some text followed by two new lines"
    replaced_content = normalize_line_endings(test_content)
    assert replaced_content == expected_content


def test_normalize_line_ending_no_new_line():
    test_content = b"This is some text without new line at the end of the file"
    expected_content = test_content
    replaced_content = normalize_line_endings(test_content)
    assert replaced_content == expected_content



import pytest
from taf.tests import TEST_WITH_REAL_YK
from taf.yubikey import (
    DEFAULT_PIN,
    export_piv_pub_key,
    export_piv_x509,
    get_serial_num,
    is_inserted,
    sign_piv_rsa_pkcs1v15,
)


@pytest.mark.skipif(not TEST_WITH_REAL_YK, reason="list_devices() is not mocked.")
def test_is_inserted():
    assert is_inserted() is True


def test_serial_num():
    assert get_serial_num() is not None


def test_export_piv_x509():
    x509_pem = export_piv_x509()
    assert isinstance(x509_pem, bytes)


def test_export_piv_pub_key():
    pub_key_pem = export_piv_pub_key()
    assert isinstance(pub_key_pem, bytes)


def test_sign_piv_rsa_pkcs1v15(targets_yk):
    targets_yk.insert()
    # yubikey-manager only supports rsa-pkcs1v15-sha256 signature scheme
    # so skip test otherwise
    if targets_yk.scheme == "rsassa-pss-sha256":
        pytest.skip()

    from securesystemslib.pyca_crypto_keys import verify_rsa_signature

    message = b"Message to be signed."
    scheme = "rsa-pkcs1v15-sha256"

    pub_key_pem = export_piv_pub_key().decode("utf-8")
    signature = sign_piv_rsa_pkcs1v15(message, DEFAULT_PIN)

    assert verify_rsa_signature(signature, scheme, pub_key_pem, message) is True



import datetime
import json
import os
from binascii import hexlify
from collections import defaultdict
from functools import partial
from getpass import getpass
from pathlib import Path

import securesystemslib
import tuf.repository_tool
from securesystemslib.exceptions import UnknownKeyError
from securesystemslib.interface import (
    import_rsa_privatekey_from_file,
    import_rsa_publickey_from_file,
)
from tuf.keydb import get_key
from tuf.repository_tool import (
    METADATA_DIRECTORY_NAME,
    TARGETS_DIRECTORY_NAME,
    create_new_repository,
    generate_and_write_rsa_keypair,
    generate_rsa_key,
    import_rsakey_from_pem,
)

from taf.auth_repo import AuthenticationRepo
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.git import GitRepository
from taf.log import get_logger
from taf.repository_tool import Repository, load_role_key
from taf.utils import get_pin_for

logger = get_logger(__name__)

try:
    import taf.yubikey as yk
except ImportError:
    logger.warning('"yubikey-manager" is not installed.')

# Yubikey x509 certificate expiration interval
EXPIRATION_INTERVAL = 36500
YUBIKEY_EXPIRATION_DATE = datetime.datetime.now() + datetime.timedelta(
    days=EXPIRATION_INTERVAL
)


def _update_target_repos(repo_path, targets_dir, target_repo_path):
    """Updates target repo's commit sha"""
    if not target_repo_path.is_dir() or target_repo_path == repo_path:
        return

    target_repo = GitRepository(str(target_repo_path))
    if target_repo.is_git_repository:
        commit = target_repo.head_commit_sha()
        target_repo_name = target_repo_path.name
        (targets_dir / target_repo_name).write_text(
            json.dumps({"commit": commit}, indent=4)
        )


def update_target_repos_from_fs(repo_path, targets_directory, namespace=None):
    """
    <Purpose>
        Create or update target files by reading the latest commits of the provided target repositories
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
    """
    repo_path = Path(repo_path).resolve()
    targets_dir = Path(targets_directory).resolve()
    if namespace is None:
        namespace = targets_dir.name
    auth_repo_targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    if namespace:
        auth_repo_targets_dir = auth_repo_targets_dir / namespace
        auth_repo_targets_dir.mkdir(parents=True, exist_ok=True)

    for target_repo_path in targets_dir.glob("*"):
        _update_target_repos(repo_path, auth_repo_targets_dir, target_repo_path)


def update_target_repos_from_repositories_json(repo_path):
    """
    <Purpose>
        Create or update target files by reading the latest commits repositories.json
    <Arguments>
        repo_path:
        Authentication repository's location
    """
    repo_path = Path(repo_path).resolve()
    targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    repositories_json = json.loads(Path(targets_dir).read_text())

    for repo in repositories_json.get("repositories"):
        target_repo_path = targets_dir / repo
        target_repo_path.mkdir(parents=True, exist_ok=True)
        _update_target_repos(repo_path, targets_dir, target_repo_path)


def build_auth_repo(
    repo_path,
    targets_directory,
    namespace,
    targets_relative_dir,
    keystore,
    roles_key_infos,
    repos_custom,
):
    # read the key infos here, no need to read the file multiple times
    roles_key_infos = _read_input_dict(roles_key_infos)
    create_repository(repo_path, keystore, roles_key_infos)
    generate_repositories_json(
        repo_path, targets_directory, namespace, targets_relative_dir, repos_custom
    )
    register_target_files(
        repo_path, keystore, roles_key_infos, commit_msg="Added repositories.json"
    )
    auth_repo_targets_dir = Path(repo_path, TARGETS_DIRECTORY_NAME)
    if namespace:
        auth_repo_targets_dir = auth_repo_targets_dir / namespace
        auth_repo_targets_dir.mkdir(parents=True, exist_ok=True)
    # group commits by dates
    # first add first repos at a date, then second repost at that date
    commits_by_date = defaultdict(dict)
    target_repositories = []
    for target_repo_dir in os.listdir(targets_directory):
        target_repo = GitRepository(Path(targets_directory, target_repo_dir))
        target_repo.checkout_branch("master")
        target_repo_name = str(Path(target_repo_dir).parent)
        target_repositories.append(target_repo_name)
        commits = target_repo.list_commits(format="format:%H|%cd", date="short")
        for commit in commits[::-1]:
            sha, date = commit.split("|")
            commits_by_date[date].setdefault(target_repo_name, []).append(sha)

    for date in sorted(commits_by_date.keys()):
        repos_and_commits = commits_by_date[date]
        for target_repo_name in target_repositories:
            if target_repo_name in repos_and_commits:
                for sha in commits_by_date[date][target_repo_name]:
                    Path(auth_repo_targets_dir, target_repo_name).write_text(
                        json.dumps({"commit": sha}, indent=4)
                    )
                    register_target_files(
                        repo_path,
                        keystore,
                        roles_key_infos,
                        commit_msg=f"Updated {target_repo_name}",
                    )


def _register_yubikey(yubikeys, role_obj, role_name, key_name, scheme, certs_dir):
    print(f"Generating keys for {key_name}")
    use_existing = False
    if len(yubikeys) > 1 or (len(yubikeys) == 1 and role_name not in yubikeys):
        use_existing = input("Do you want to reuse already set up Yubikey? y/n ") == "y"
        if use_existing:
            existing_key = None
            key_id_certs = {}
            while existing_key is None:
                for existing_role_name, role_keys in yubikeys.items():
                    if existing_role_name == role_name:
                        continue
                    print(f"Existing keys for role {existing_role_name} are:\n")
                    for key_and_cert in role_keys.values():
                        key, cert_cn = key_and_cert
                        key_id_certs[key["keyid"]] = cert_cn
                        print(f"{cert_cn} id: {key['keyid']}")
                existing_keyid = input("\nEnter existing Yubikey's id and press ENTER ")
                try:
                    existing_key = get_key(existing_keyid)
                    cert_cn = key_id_certs[existing_keyid]
                except UnknownKeyError:
                    pass
    if not use_existing:
        input("Please insert a new YubiKey and press ENTER.")
        serial_num = yk.get_serial_num()
        while serial_num in yubikeys[role_name]:
            print(f"Yubikey with serial number {serial_num} is already in use.\n")
            input("Please insert new YubiKey and press ENTER.")
            serial_num = yk.get_serial_num()

        pin = get_pin_for(key_name)

        cert_cn = input("Enter key holder's name: ")

        print("Generating keys, please wait...")
        pub_key_pem = yk.setup(pin, cert_cn, cert_exp_days=EXPIRATION_INTERVAL).decode(
            "utf-8"
        )

        key = import_rsakey_from_pem(pub_key_pem, scheme)

        cert_path = Path(certs_dir, f"{key['keyid']}.cert")
        with open(cert_path, "wb") as f:
            f.write(yk.export_piv_x509())

    # set Yubikey expiration date
    role_obj.add_verification_key(key, expires=YUBIKEY_EXPIRATION_DATE)
    role_obj.add_external_signature_provider(
        key, partial(signature_provider, key["keyid"], cert_cn)
    )
    yubikeys[role_name][serial_num] = (key, cert_cn)


def _register_key(keystore, passwords, role_obj, role_name, key_name, key_num, scheme):
    # if keystore exists, load the keys
    # this is useful when generating tests
    if keystore is not None:
        public_key = import_rsa_publickey_from_file(
            str(Path(keystore, f"{key_name}.pub")), scheme
        )
        password = passwords[key_num]
        if password:
            private_key = import_rsa_privatekey_from_file(
                str(Path(keystore, key_name)), password, scheme=scheme
            )
        else:
            private_key = import_rsa_privatekey_from_file(
                str(Path(keystore, key_name)), scheme=scheme
            )

    # if it does not, generate the keys and print the output
    else:
        key = generate_rsa_key()
        print(f"{role_name} key:\n\n{key['keyval']['private']}\n\n")
        public_key = private_key = key
    role_obj.add_verification_key(public_key)
    role_obj.load_signing_key(private_key)


def create_repository(
    repo_path, keystore, roles_key_infos, commit_message=None, test=False
):
    """
    <Purpose>
        Create a new authentication repository. Generate initial metadata files.
        The initial targets metadata file is empty (does not specify any targets).
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_message:
        If provided, the changes will be committed automatically using the specified message
        test:
        Indicates if the created repository is a test authentication repository
    """
    yubikeys = defaultdict(dict)
    roles_key_infos = _read_input_dict(roles_key_infos)
    repo = AuthenticationRepo(repo_path)
    if Path(repo_path).is_dir():
        if repo.is_git_repository:
            print(f"Repository {repo_path} already exists")
            return

    tuf.repository_tool.METADATA_STAGED_DIRECTORY_NAME = METADATA_DIRECTORY_NAME
    repository = create_new_repository(repo_path)
    for role_name, key_info in roles_key_infos.items():
        num_of_keys = key_info.get("number", 1)
        passwords = key_info.get("passwords", [None] * num_of_keys)
        threshold = key_info.get("threshold", 1)
        is_yubikey = key_info.get("yubikey", False)
        scheme = key_info.get("scheme", DEFAULT_RSA_SIGNATURE_SCHEME)

        role_obj = _role_obj(role_name, repository)
        role_obj.threshold = threshold
        for key_num in range(num_of_keys):
            key_name = _get_key_name(role_name, key_num, num_of_keys)
            if is_yubikey:
                _register_yubikey(
                    yubikeys, role_obj, role_name, key_name, scheme, repo.certs_dir
                )
            else:
                _register_key(
                    keystore, passwords, role_obj, role_name, key_name, key_num, scheme
                )

    # if the repository is a test repository, add a target file called test-auth-repo
    if test:
        target_paths = Path(repo_path) / "targets"
        test_auth_file = target_paths / "test-auth-repo"
        test_auth_file.touch()
        targets_obj = _role_obj("targets", repository)
        targets_obj.add_target(str(test_auth_file))

    repository.writeall()
    if commit_message is not None and len(commit_message):
        auth_repo = GitRepository(repo_path)
        auth_repo.init_repo()
        auth_repo.commit(commit_message)


def generate_keys(keystore, roles_key_infos):
    """
    <Purpose>
        Generate public and private keys and writes them to disk. Names of keys correspond to names
        of the TUF roles. If more than one key should be generated per role, a counter is appended
        to the role's name. E.g. root1, root2, root3 etc.
    <Arguments>
        keystore:
        Location where the generated files should be saved
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        This includes:
            - passwords of the keystore files
            - number of keys per role (optional, defaults to one if not provided)
            - key length (optional, defaults to TUF's default value, which is 3072)
        Names of the keys are set to names of the roles plus a counter, if more than one key
        should be generated.
    """
    roles_key_infos = _read_input_dict(roles_key_infos)
    for role_name, key_info in roles_key_infos.items():
        num_of_keys = key_info.get("number", 1)
        bits = key_info.get("length", 3072)
        passwords = key_info.get("passwords", [""] * num_of_keys)
        is_yubikey = key_info.get("yubikey", False)
        for key_num in range(num_of_keys):
            if not is_yubikey:
                key_name = _get_key_name(role_name, key_num, num_of_keys)
                password = passwords[key_num]
                generate_and_write_rsa_keypair(
                    str(Path(keystore, key_name)), bits=bits, password=password
                )


def generate_repositories_json(
    repo_path,
    targets_directory,
    namespace=None,
    targets_relative_dir=None,
    custom_data=None,
):
    """
    <Purpose>
        Generatesinitial repositories.json
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
        targets_relative_dir:
        Directory relative to which urls of the target repositories are set, if they do not have remote set
    """

    custom_data = _read_input_dict(custom_data)
    repositories = {}

    repo_path = Path(repo_path).resolve()
    auth_repo_targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    targets_directory = Path(targets_directory).resolve()
    if targets_relative_dir is not None:
        targets_relative_dir = Path(targets_relative_dir).resolve()
    if namespace is None:
        namespace = targets_directory.name
    for target_repo_dir in targets_directory.glob("*"):
        if not target_repo_dir.is_dir() or target_repo_dir == repo_path:
            continue
        target_repo = GitRepository(target_repo_dir.resolve())
        if not target_repo.is_git_repository:
            continue
        target_repo_name = target_repo_dir.name
        target_repo_namespaced_name = (
            target_repo_name if not namespace else f"{namespace}/{target_repo_name}"
        )
        # determine url to specify in initial repositories.json
        # if the repository has a remote set, use that url
        # otherwise, set url to the repository's absolute or relative path (relative
        # to targets_relative_dir if it is specified)
        url = target_repo.get_remote_url()
        if url is None:
            if targets_relative_dir is not None:
                url = Path(target_repo.repo_path).relative_to(targets_relative_dir)
            else:
                url = Path(target_repo.repo_path).resolve()
            # convert to posix path
            url = str(url.as_posix())
        repositories[target_repo_namespaced_name] = {"urls": [url]}
        if target_repo_namespaced_name in custom_data:
            repositories[target_repo_namespaced_name]["custom"] = custom_data[
                target_repo_namespaced_name
            ]

    (auth_repo_targets_dir / "repositories.json").write_text(
        json.dumps({"repositories": repositories}, indent=4)
    )


def _get_key_name(role_name, key_num, num_of_keys):
    if num_of_keys == 1:
        return role_name
    else:
        return role_name + str(key_num + 1)


def init_repo(
    repo_path,
    targets_directory,
    namespace,
    targets_relative_dir,
    keystore,
    roles_key_infos,
    repos_custom=None,
    commit=None,
    test=False,
):
    """
    <Purpose>
        Generate initial repository:
        1. Crete tuf authentication repository
        2. Commit initial metadata files if commit == True
        3. Add target repositories
        4. Generate repositories.json
        5. Update tuf metadata
        6. Commit the changes if commit == True
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
        targets_relative_dir:
        Directory relative to which urls of the target repositories are set, if they do not have remote set
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_message:
        If provided, the changes will be committed automatically using the specified message
        test:
        Indicates if the created repository is a test authentication repository
    """
    # read the key infos here, no need to read the file multiple times
    roles_key_infos = _read_input_dict(roles_key_infos)
    commit_msg = "Initial commit" if commit else None
    create_repository(repo_path, keystore, roles_key_infos, commit_msg, test)
    update_target_repos_from_fs(repo_path, targets_directory, namespace)
    generate_repositories_json(
        repo_path, targets_directory, namespace, targets_relative_dir, repos_custom
    )
    register_target_files(repo_path, keystore, roles_key_infos, commit_msg=commit)


def _load_role_key_from_keys_dict(role, roles_key_infos):
    password = None
    if roles_key_infos is not None and len(roles_key_infos):
        if role in roles_key_infos:
            password = roles_key_infos[role].get("passwords", [None])[0] or None
    return password


def register_target_file(repo_path, file_path, keystore, roles_key_infos, scheme):
    roles_key_infos = _read_input_dict(roles_key_infos)
    taf_repo = Repository(repo_path)
    taf_repo.add_existing_target(file_path)

    _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme)


def _read_input_dict(value):
    if value is None:
        return {}
    if type(value) is str:
        if Path(value).is_file():
            with open(value) as f:
                value = json.loads(f.read())
        else:
            value = json.loads(value)
    return value


def register_target_files(
    repo_path,
    keystore,
    roles_key_infos,
    commit_msg=None,
    scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
):
    """
    <Purpose>
        Register all files found in the target directory as targets - updates the targets
        metadata file, snapshot and timestamp. Sign targets
        with yubikey if keystore is not provided
    <Arguments>
        repo_path:
        Authentication repository's path
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_msg:
        Commit message. If specified, the changes made to the authentication are committed.
        scheme:
        A signature scheme used for signing.
    """
    roles_key_infos = _read_input_dict(roles_key_infos)
    repo_path = Path(repo_path).resolve()
    targets_path = repo_path / TARGETS_DIRECTORY_NAME
    taf_repo = Repository(str(repo_path))
    for root, _, filenames in os.walk(str(targets_path)):
        for filename in filenames:
            taf_repo.add_existing_target(str(Path(root) / filename))
    _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme)
    if commit_msg is not None:
        auth_repo = GitRepository(repo_path)
        auth_repo.commit(commit_msg)


def _role_obj(role, repository):
    if role == "targets":
        return repository.targets
    elif role == "snapshot":
        return repository.snapshot
    elif role == "timestamp":
        return repository.timestamp
    elif role == "root":
        return repository.root


def signature_provider(key_id, cert_cn, key, data):  # pylint: disable=W0613
    def _check_key_id(expected_key_id):
        try:
            inserted_key = yk.get_piv_public_key_tuf()
            return expected_key_id == inserted_key["keyid"]
        except Exception:
            return False

    while not _check_key_id(key_id):
        pass

    data = securesystemslib.formats.encode_canonical(data).encode("utf-8")
    key_pin = getpass(f"Please insert {cert_cn} YubiKey, input PIN and press ENTER.\n")
    signature = yk.sign_piv_rsa_pkcs1v15(data, key_pin)

    return {"keyid": key_id, "sig": hexlify(signature).decode()}


def update_metadata_expiration_date(
    repo_path,
    keystore,
    roles_key_infos,
    role,
    start_date=datetime.datetime.now(),
    interval=None,
    commit_msg=None,
):
    roles_key_infos = _read_input_dict(roles_key_infos)
    taf_repo = Repository(repo_path)
    update_methods = {
        "timestamp": taf_repo.update_timestamp,
        "snapshot": taf_repo.update_snapshot,
        "targets": taf_repo.update_targets_from_keystore,
    }
    password = _load_role_key_from_keys_dict(role, roles_key_infos)
    update_methods[role](keystore, password, start_date, interval)

    if commit_msg is not None:
        auth_repo = GitRepository(repo_path)
        auth_repo.commit(commit_msg)


def _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme):
    if keystore is not None:
        # load all keys from keystore files
        # convenient when generating test repositories
        # not recommended in production
        targets_password = _load_role_key_from_keys_dict("targets", roles_key_infos)
        targets_key = load_role_key(keystore, "targets", targets_password, scheme)
        taf_repo.update_targets_from_keystore(targets_key, write=False)
        snapshot_password = _load_role_key_from_keys_dict("snapshot", roles_key_infos)
        timestamp_password = _load_role_key_from_keys_dict("timestamp", roles_key_infos)
        timestamp_key = load_role_key(keystore, "timestamp", timestamp_password)
        snapshot_key = load_role_key(keystore, "snapshot", snapshot_password)
    else:
        targets_key_pin = getpass(
            "Please insert targets YubiKey, input PIN and press ENTER."
        )
        taf_repo.update_targets(targets_key_pin, write=False)
        snapshot_pem = getpass("Enter snapshot key")
        snapshot_pem = _form_private_pem(snapshot_pem)
        snapshot_key = import_rsakey_from_pem(snapshot_pem, scheme)

        timestamp_pem = getpass("Enter timestamp key")
        timestamp_pem = _form_private_pem(timestamp_pem)
        timestamp_key = import_rsakey_from_pem(timestamp_pem, scheme)
    taf_repo.update_snapshot_and_timestmap(snapshot_key, timestamp_key, write=False)
    taf_repo.writeall()


def _form_private_pem(pem):
    return f"-----BEGIN RSA PRIVATE KEY-----\n{pem}\n-----END RSA PRIVATE KEY-----"


# TODO Implement update of repositories.json (updating urls, custom data, adding new repository, removing
# repository etc.)
# TODO create tests for this



import json
import os
import re
import shutil
import subprocess
from collections import OrderedDict
from pathlib import Path

import taf.log
import taf.settings as settings
from taf.exceptions import InvalidRepositoryError
from taf.utils import run

logger = taf.log.get_logger(__name__)

EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"


class GitRepository(object):
    def __init__(
        self, repo_path, repo_urls=None, additional_info=None, default_branch="master"
    ):
        """
    Args:
      repo_path: repository's path
      repo_urls: repository's urls (optional)
      additional_info: a dictionary containing other data (optional)
      default_branch: repository's default branch
    """
        self.repo_path = str(repo_path)
        self.default_branch = default_branch
        if repo_urls is not None:
            if settings.update_from_filesystem is False:
                for url in repo_urls:
                    _validate_url(url)
            else:
                repo_urls = [
                    os.path.normpath(os.path.join(self.repo_path, url))
                    if not os.path.isabs(url)
                    else url
                    for url in repo_urls
                ]
        self.repo_urls = repo_urls
        self.additional_info = additional_info
        self.repo_name = os.path.basename(self.repo_path)

    _remotes = None

    @property
    def remotes(self):
        if self._remotes is None:
            self._remotes = self._git("remote").split("\n")
        return self._remotes

    @property
    def is_git_repository_root(self):
        """Check if path is git repository."""
        git_path = Path(self.repo_path) / ".git"
        return self.is_git_repository and (git_path.is_dir() or git_path.is_file())

    @property
    def is_git_repository(self):
        try:
            self._git("rev-parse --git-dir")
            return True
        except subprocess.CalledProcessError:
            return False

    @property
    def initial_commit(self):
        return (
            self._git("rev-list --max-parents=0 HEAD").strip()
            if self.is_git_repository
            else None
        )

    def is_remote_branch(self, branch_name):
        for remote in self.remotes:
            if branch_name.startswith(remote + "/"):
                return True
        return False

    def _git(self, cmd, *args, **kwargs):
        """Call git commands in subprocess
    e.g.:
      self._git('checkout {}', branch_name)
    """
        log_error = kwargs.pop("log_error", False)
        log_error_msg = kwargs.pop("log_error_msg", "")
        reraise_error = kwargs.pop("reraise_error", False)
        log_success_msg = kwargs.pop("log_success_msg", "")

        if len(args):
            cmd = cmd.format(*args)
        command = f"git -C {self.repo_path} {cmd}"
        if log_error or log_error_msg:
            try:
                result = run(command)
                if log_success_msg:
                    logger.debug("Repo %s:" + log_success_msg, self.repo_name)
            except subprocess.CalledProcessError as e:
                if log_error_msg:
                    logger.error(log_error_msg)
                else:
                    logger.error(
                        "Repo %s: error occurred while executing %s:\n%s",
                        self.repo_name,
                        command,
                        e.output,
                    )
                if reraise_error:
                    raise
        else:
            result = run(command)
            if log_success_msg:
                logger.debug("Repo %s: " + log_success_msg, self.repo_name)
        return result

    def all_commits_on_branch(self, branch=None, reverse=True):
        """Returns a list of all commits on the specified branch. If branch is None,
        all commits on the currently checked out branch will be returned
        """
        if branch is None:
            branch = ""
        commits = self._git("log {} --format=format:%H", branch).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            if reverse:
                commits.reverse()

        logger.debug(
            "Repo %s: found the following commits: %s",
            self.repo_name,
            ", ".join(commits),
        )
        return commits

    def all_commits_since_commit(self, since_commit, reverse=True):
        """Returns a list of all commits since the specified commit on the
        currently checked out branch
        """
        if since_commit is None:
            return self.all_commits_on_branch(reverse=reverse)
        commits = self._git("rev-list {}..HEAD", since_commit).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            if reverse:
                commits.reverse()

        logger.debug(
            "Repo %s: found the following commits after commit %s: %s",
            self.repo_name,
            since_commit,
            ", ".join(commits),
        )
        return commits

    def all_fetched_commits(self, branch="master"):
        commits = self._git("rev-list ..origin/{}", branch).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            commits.reverse()
        logger.debug(
            "Repo %s: fetched the following commits %s",
            self.repo_name,
            ", ".join(commits),
        )
        return commits

    def branches(self):
        """Returns all branches."""
        return [
            branch.strip('"').strip("'").strip()
            for branch in self._git("branch --format='%(refname:short)'").split("\n")
        ]

    def branches_containing_commit(self, commit, strip_remote=False):
        """Finds all branches that contain the given commit"""
        local_branches = self._git(f"branch --contains {commit}").split("\n")
        if local_branches:
            local_branches = [
                branch.replace("*", "").strip() for branch in local_branches
            ]
        remote_branches = self._git(f"branch -r --contains {commit}").split("\n")
        filtered_remote_branches = []
        if remote_branches:
            for branch in remote_branches:
                branch = branch.strip()
                local_name = self.branch_local_name(branch)
                if (
                    local_name
                    and "HEAD" not in local_name
                    and local_name not in local_branches
                ):
                    if strip_remote:
                        branch = self.branch_local_name(branch)
                    filtered_remote_branches.append(branch)
        branches = {branch: False for branch in local_branches}
        branches.update({branch: True for branch in filtered_remote_branches})
        return OrderedDict(sorted(branches.items(), reverse=True))

    def branch_exists(self, branch_name, include_remotes=True):
        """
        Checks if a branch with the given name exists.
        If include_remotes is set to True, this checks if
        a remote branch exists.
        """
        branch = self._git(f"branch --list {branch_name}")
        if branch:
            return True
        if include_remotes:
            for remote in self.remotes:
                branch = self._git(f"branch -r --list origin/{branch_name}")
                if branch:
                    return True
        return False

    def branch_off_commit(self, branch_name, commit):
        """Create a new branch by branching off of the specified commit"""
        self._git(f"checkout -b {branch_name} {commit}")

    def branch_local_name(self, remote_branch_name):
        """Strip remote from the given remote branch"""
        for remote in self.remotes:
            if remote_branch_name.startswith(remote + "/"):
                return remote_branch_name.split("/", 1)[1]

    def checkout_branch(self, branch_name, create=False):
        """Check out the specified branch. If it does not exists and
    the create parameter is set to True, create a new branch.
    If the branch does not exist and create is set to False,
    raise an exception."""
        try:
            self._git(
                "checkout {}",
                branch_name,
                log_error=True,
                reraise_error=True,
                log_success_msg=f"checked out branch {branch_name}",
            )
        except subprocess.CalledProcessError as e:
            if create:
                self.create_and_checkout_branch(branch_name)
            else:
                raise (e)

    def checkout_paths(self, commit, *args):
        for file_path in args:
            self._git(f"checkout {commit} {file_path}")

    def clean(self):
        self._git("clean -fd")

    def clone(self, no_checkout=False, bare=False):

        logger.info("Repo %s: cloning repository", self.repo_name)
        shutil.rmtree(self.repo_path, True)
        os.makedirs(self.repo_path, exist_ok=True)
        if self.repo_urls is None:
            raise Exception("Cannot clone repository. No urls were specified")
        params = ""
        if bare:
            params = "--bare"
        elif no_checkout:
            params = "--no-checkout"
        for url in self.repo_urls:
            try:
                self._git(
                    "clone {} . {}", url, params, log_success_msg="successfully cloned"
                )
            except subprocess.CalledProcessError:
                logger.error("Repo %s: cannot clone from url %s", self.repo_name, url)
            else:
                break

    def create_and_checkout_branch(self, branch_name, raise_error_if_exists=True):
        flag = "-b" if raise_error_if_exists else "-B"
        self._git(
            "checkout {} {}",
            flag,
            branch_name,
            log_success_msg=f"created and checked out branch {branch_name}",
            log_error=True,
            reraise_error=True,
        )

    def create_branch(self, branch_name, commit=None):
        if commit is None:
            commit = ""
        self._git(
            "branch {} {}",
            branch_name,
            commit,
            log_success_msg=f"created branch {branch_name}",
            log_error=True,
            reraise_error=True,
        )

    def create_local_branch(self, branch_name):
        """Create local branch by checking it if it does not exist out and making sure
        to check out previously checked out branch
        """
        if not self.branch_exists(branch_name, include_remotes=False):
            current_branch = self.get_current_branch()
            self.checkout_branch(branch_name)
            self.checkout_branch(current_branch)

    def checkout_commit(self, commit):
        self._git("checkout {}", commit, log_success_msg=f"checked out commit {commit}")

    def commit(self, message):
        """Create a commit with the provided message on the currently checked out branch"""
        self._git("add -A")
        try:
            self._git("diff --cached --exit-code --shortstat")
        except subprocess.CalledProcessError:
            run("git", "-C", self.repo_path, "commit", "--quiet", "-m", message)
        return self._git("rev-parse HEAD")

    def commit_empty(self, message):
        run(
            "git",
            "-C",
            self.repo_path,
            "commit",
            "--quiet",
            "--allow-empty",
            "-m",
            message,
        )
        return self._git("rev-parse HEAD")

    def commits_on_branch_and_not_other(
        self, branch1, branch2, include_branching_commit=False
    ):
        """
        Meant to find commits belonging to a branch which branches off of
        a commit from another branch. For example, to find only commits
        on a speculative branch and not on the master branch.
        """

        logger.debug(
            "Repo %s: finding commits which are on branch %s, but not on branch %s",
            self.repo_name,
            branch1,
            branch2,
        )
        commits = self._git(
            "log {} --not {} --no-merges --format=format:%H", branch1, branch2
        )
        commits = commits.split("\n") if commits else []
        if include_branching_commit:
            branching_commit = self._git("rev-list -n 1 {}~1", commits[-1])
            commits.append(branching_commit)
        logger.debug(
            "Repo %s: found the following commits: %s", self.repo_name, commits
        )
        return commits

    def delete_local_branch(self, branch_name, force=False):
        """Deletes local branch."""
        flag = "-D" if force else "-d"
        self._git(f"branch {flag} {branch_name}")

    def delete_remote_branch(self, branch_name, remote=None):
        if remote is None:
            remote = self.remotes[0]
        self._git(f"push {remote} --delete {branch_name}")

    def get_commits_date(self, commit):
        date = self._git("show -s --format=%at {}", commit)
        return date.split(" ", 1)[0]

    def get_commit_message(self, commit):
        """Returns commit message of the given commit"""
        return self._git("log --format=%B -n 1 {}", commit).strip()

    def get_commit_sha(self, behind_head):
        """Get commit sha of HEAD~{behind_head}"""
        return self._git("rev-parse HEAD~{}", behind_head)

    def get_json(self, commit, path):
        s = self.get_file(commit, path)
        return json.loads(s)

    def get_file(self, commit, path):
        return self._git("show {}:{}", commit, path)

    def get_last_branch_by_committer_date(self):
        """Find the latest branch based on committer date. Should only be used for
        testing purposes"""
        branches = self._git("branch --sort=committerdate").strip().split("\n")
        if not len(branches):
            return None
        return branches[-1]

    def get_remote_url(self):
        try:
            return self._git("config --get remote.origin.url").strip()
        except subprocess.CalledProcessError:
            return None

    def delete_branch(self, branch_name):
        self._git("branch -D {}", branch_name)

    def diff_between_revisions(self, revision1=EMPTY_TREE, revision2="HEAD"):
        return self._git("diff {} {}", revision1, revision2)

    def has_remote(self):
        return bool(self._git("remote"))

    def head_commit_sha(self):
        """Finds sha of the commit to which the current HEAD points"""
        try:
            return self._git("rev-parse HEAD")
        except subprocess.CalledProcessError:
            return None

    def fetch(self, fetch_all=False):
        if fetch_all:
            self._git("fetch --all")
        else:
            self._git("fetch")

    def get_current_branch(self):
        """Return current branch."""
        return self._git("rev-parse --abbrev-ref HEAD").strip()

    def get_merge_base(self, branch1, branch2):
        """Finds the best common ancestor between two branches"""
        return self._git(f"merge-base {branch1} {branch2}")

    def get_tracking_branch(self, branch=""):
        """Returns tracking branch name in format origin/branch-name or None if branch does not
        track remote branch.
        """
        try:
            return self._git(f"rev-parse --abbrev-ref {branch}@{{u}}")
        except subprocess.CalledProcessError:
            return None

    def init_repo(self, bare=False):
        if not os.path.isdir(self.repo_path):
            os.makedirs(self.repo_path, exist_ok=True)
        flag = "--bare" if bare else ""
        self._git(f"init {flag}")
        if self.repo_urls is not None and len(self.repo_urls):
            self._git("remote add origin {}", self.repo_urls[0])

    def list_files_at_revision(self, commit, path=""):
        if path is None:
            path = ""
        file_names = self._git("ls-tree -r --name-only {}", commit)
        list_of_files = []
        if not file_names:
            return list_of_files
        for file_in_repo in file_names.split("\n"):
            if not file_in_repo.startswith(path):
                continue
            file_in_repo = os.path.relpath(file_in_repo, path)
            list_of_files.append(file_in_repo)
        return list_of_files

    def list_commits(self, **kwargs):
        params = []
        for name, value in kwargs.items():
            params.append(f"--{name}={value}")

        return self._git("log {}", " ".join(params)).split("\n")

    def list_n_commits(self, number=10, skip=None, branch=None):
        """List the specified number of top commits of the current branch
        Optionally skip a number of commits from the top"""
        if branch is None:
            branch = ""
        if skip:
            commits = self._git(
                f"log {branch} --format=format:%H --skip={skip} -n {number}"
            )
        else:
            commits = self._git(f"log {branch} --format=format:%H -n {number}")
        return commits.split("\n") if commits else []

    def merge_commit(self, commit):
        self._git("merge {}", commit)

    def pull(self):
        """Pull current branch"""
        self._git("pull")

    def push(self, branch=None, set_upstream=False, force=False):
        """Push all changes"""
        if branch is None:
            branch = self.get_current_branch()
        upstream_flag = "-u" if set_upstream else ""
        force_flag = "-f" if force else ""
        self._git("push {} {} origin {}", upstream_flag, force_flag, branch)

    def rename_branch(self, old_name, new_name):
        self._git("branch -m {} {}", old_name, new_name)

    def reset_num_of_commits(self, num_of_commits, hard=False):
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} HEAD~{num_of_commits}")

    def reset_to_commit(self, commit, hard=False):
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} {commit}")

    def reset_to_head(self):
        self._git("reset --hard HEAD")

    def set_upstream(self, branch_name):
        self._git("branch -u origin/{}", branch_name)

    def something_to_commit(self):
        """Checks if there are any uncommitted changes"""
        uncommitted_changes = self._git("status --porcelain")
        return bool(uncommitted_changes)

    def synced_with_remote(self, branch="master"):
        """Checks if local branch is synced with its remote branch"""
        # check if the latest local commit matches
        # the latest remote commit on the specified branch
        tracking_branch = self.get_tracking_branch(branch)
        if not tracking_branch:
            return False

        try:
            local_commit = self._git(f"rev-parse {branch}")
        except subprocess.CalledProcessError as e:
            if "unknown revision or path not in the working tree" not in e.output:
                raise e
            local_commit = None

        try:
            remote_commit = self._git(f"rev-parse {tracking_branch}")
        except subprocess.CalledProcessError as e:
            if "unknown revision or path not in the working tree" not in e.output:
                raise e
            remote_commit = None

        return local_commit == remote_commit

    def top_commit_of_branch(self, branch):
        return self._git(f"rev-parse {branch}")


class NamedGitRepository(GitRepository):
    def __init__(
        self,
        root_dir,
        repo_name,
        repo_urls=None,
        additional_info=None,
        default_branch="master",
    ):
        """
    Args:
      root_dir: the root directory
      repo_name: repository's path relative to the root directory root_dir
      repo_urls: repository's urls (optional)
      additional_info: a dictionary containing other data (optional)
      default_branch: repository's default branch
    repo_path is the absolute path to this repository. It is set by joining
    root_dir and repo_name.
    """
        repo_path = _get_repo_path(root_dir, repo_name)
        super().__init__(repo_path, repo_urls, additional_info, default_branch)
        self.repo_name = repo_name


def _get_repo_path(root_dir, repo_name):
    """
  get the path to a repo and ensure it is valid.
  (since this is coming from potentially untrusted data)
  """
    _validate_repo_name(repo_name)
    repo_dir = str((Path(root_dir) / (repo_name or "")))
    if not repo_dir.startswith(repo_dir):
        logger.error("Repo %s: repository name is not valid", repo_name)
        raise InvalidRepositoryError(f"Invalid repository name: {repo_name}")
    return repo_dir


_repo_name_re = re.compile(r"^\w[\w_-]*/\w[\w_-]*$")


def _validate_repo_name(repo_name):
    """ Ensure the repo name is not malicious """
    match = _repo_name_re.match(repo_name)
    if not match:
        logger.error("Repo %s: repository name is not valid", repo_name)
        raise InvalidRepositoryError(
            "Repository name must be in format namespace/repository "
            "and can only contain letters, numbers, underscores and "
            f'dashes, but got "{repo_name}"'
        )


_http_fttp_url = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    # domain...
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

_ssh_url = re.compile(
    r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)?(/)?"
)


def _validate_url(url):
    """ ensure valid URL """
    for _url_re in [_http_fttp_url, _ssh_url]:
        match = _url_re.match(url)
        if match:
            return
    logger.error("Repository URL (%s) is not valid", url)
    raise InvalidRepositoryError(
        f'Repository URL must be a valid URL, but got "{url}".'
    )



import logging
import os
from pathlib import Path

import taf.settings

_FORMAT_STRING = (
    "[%(asctime)s] [%(levelname)s] "
    + "[%(funcName)s:%(lineno)s@%(filename)s]\n%(message)s\n"
)
formatter = logging.Formatter(_FORMAT_STRING)

logger = logging.getLogger("taf")
logger.setLevel(taf.settings.LOG_LEVEL)


def _get_log_location():
    location = taf.settings.LOGS_LOCATION or os.environ.get("TAF_LOG")
    if location is None:
        location = Path.home() / ".taf"
        location.mkdir(exist_ok=True)
    else:
        location = Path(location)
    return location


if taf.settings.ENABLE_CONSOLE_LOGGING:
    console_handler = logging.StreamHandler()
    console_handler.setLevel(taf.settings.CONSOLE_LOGGING_LEVEL)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)


if taf.settings.ENABLE_FILE_LOGGING:
    logs_location = _get_log_location()
    file_handler = logging.FileHandler(str(logs_location / taf.settings.LOG_FILENAME))
    file_handler.setLevel(taf.settings.FILE_LOGGING_LEVEL)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    if taf.settings.SEPARATE_ERRORS:
        error_handler = logging.FileHandler(
            str(logs_location / taf.settings.ERROR_LOG_FILENAME)
        )
        error_handler.setLevel(taf.settings.ERROR_LOGGING_LEVEL)
        error_handler.setFormatter(formatter)
        logger.addHandler(error_handler)


def get_logger(name):
    return logging.getLogger(name)



import datetime
from pathlib import Path

import click

import taf.developer_tool as developer_tool
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.updater.updater import update_named_repository, update_repository
from taf.utils import ISO_DATE_PARAM_TYPE as ISO_DATE


@click.group()
def cli():
    pass


@cli.command()
@click.option('--repo-path', default='repository', help='Authentication repository\'s path')
@click.option('--keystore', default=None, help='Path of the keystore file')
@click.option('--keys-description', default=None, help='A dictionary containing information about the keys or a path'
              ' to a json file which which stores the needed information')
@click.option('--commit-msg', default=None, help='Commit message to be used in case the changes'
              'should be automatically committed')
@click.option('--scheme', default=DEFAULT_RSA_SIGNATURE_SCHEME, help='A signature scheme used for signing.')
def sign_targets(repo_path, keystore, keys_description, commit_msg, scheme):
    developer_tool.register_target_files(repo_path, keystore, keys_description, commit_msg, scheme)


@cli.command()
@click.option('--repo-path', default='repository', help='Authentication repository\'s path')
@click.option('--file-path', help="Target file's path, relative to the targets directory")
@click.option('--keystore', default='keystore', help='Path of the keystore file')
@click.option('--keys-description', default=None, help='A dictionary containing information about the keys or a path'
              ' to a json file which which stores the needed information')
@click.option('--scheme', default=DEFAULT_RSA_SIGNATURE_SCHEME, help='A signature scheme used for signing.')
def add_target_file(repo_path, file_path, keystore, keys_description, scheme):
    developer_tool.register_target_file(repo_path, file_path, keystore, keys_description, scheme)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the repository')
@click.option('--targets-dir', default='targets', help='Directory where the target '
              'repositories are located')
@click.option('--namespace', default=None, help='Namespace of the target repositories')
def update_repos_from_fs(repo_path, targets_dir, namespace):
    "Updates target repositories by traversing given targets directory "
    developer_tool.update_target_repos_from_fs(repo_path, targets_dir, namespace)


@cli.command()
@click.option("--repo-path", default="repository", help="Location of the repository")
@click.option('--targets-dir', default='targets', help='Directory where the target '
              'repositories are located')
@click.option('--namespace', default=None, help='Namespace of the target repositories')
def update_repos_from_repositories_json(repo_path, targets_dir, namespace):
    "Updates target repositories by traversing repositories.json "
    developer_tool.update_target_repos_from_repositories_json(repo_path)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the authentication repository')
@click.option('--targets-dir', default='targets', help='Directory where the target '
              'repositories are located')
@click.option('--namespace', default='', help='Namespace of the target repositories')
@click.option('--targets-rel-dir', default=None, help=' Directory relative to which urls '
              'of the target repositories are set, if they do not have remote set')
@click.option('--keystore', default='keystore', help='Location of the keystore file')
@click.option('--keys-description', help='A dictionary containing information about the '
              'keys or a path to a json file which which stores the needed information')
@click.option('--custom', default=None, help='A dictionary containing custom '
              'targets info which will be included in repositories.json')
def build_auth_repo(repo_path, targets_dir, namespace, targets_rel_dir, keystore,
                    keys_description, custom):
    developer_tool.build_auth_repo(repo_path, targets_dir, namespace, targets_rel_dir, keystore,
                                   keys_description, custom)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the repository')
@click.option('--keystore', default=None, help='Location of the keystore file')
@click.option('--keys-description', help='A dictionary containing information about the '
              'keys or a path to a json file which which stores the needed information')
@click.option('--commit-msg', default=None, help='Commit message. If provided, the '
              'changes will be committed automatically')
@click.option('--test', is_flag=True, default=False, help='Indicates if the created repository '
              'is a test authentication repository')
def create_repo(repo_path, keystore, keys_description, commit_msg, test):
    developer_tool.create_repository(repo_path, keystore, keys_description, commit_msg, test)


@cli.command()
@click.option('--keystore', default='keystore', help='Location of the keystore file')
@click.option('--keys-description', help='A dictionary containing information about the keys or a path'
              ' to a json file which which stores the needed information')
def generate_keys(keystore, keys_description):
    developer_tool.generate_keys(keystore, keys_description)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the repository')
@click.option('--targets-dir', default='targets', help='Directory where the target '
              'repositories are located')
@click.option('--namespace', default='', help='Namespace of the target repositories')
@click.option('--targets-rel-dir', default=None, help=' Directory relative to which urls '
              'of the target repositories are set, if they do not have remote set')
@click.option('--keystore', default='keystore', help='Location of the keystore file')
@click.option('--keys-description', help='A dictionary containing information about the '
              'keys or a path to a json file which which stores the needed information')
@click.option('--custom', default=None, help='A dictionary containing custom '
              'targets info which will be included in repositories.json')
@click.option('--commit', is_flag=True, default=True, help='Indicates if changes should be committed')
@click.option('--test', is_flag=True, default=False, help='Indicates if the created repository '
              'is a test authentication repository')
def init_repo(repo_path, targets_dir, namespace, targets_rel_dir, keystore,
              keys_description, custom, commit, test):
    developer_tool.init_repo(repo_path, targets_dir, namespace, targets_rel_dir, keystore,
                             keys_description, repos_custom=custom, commit=commit, test=test)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the repository')
@click.option('--targets-dir', default='targets', help='Directory where the target '
              'repositories are located')
@click.option('--namespace', default=None, help='Namespace of the target repositories')
@click.option('--targets-rel-dir', default=None, help=' Directory relative to which urls '
              'of the target repositories are set, if they do not have remote set')
@click.option('--custom', default=None, help='A dictionary containing custom '
              'targets info which will be included in repositories.json')
def generate_repositories_json(repo_path, targets_dir, namespace, targets_rel_dir, custom):
    developer_tool.generate_repositories_json(repo_path, targets_dir, namespace, targets_rel_dir,
                                              custom)


@cli.command()
@click.option('--repo-path', default='repository', help='Location of the repository')
@click.option('--keystore', default='keystore', help='Location of the keystore file')
@click.option('--keys-description', help='A dictionary containing information about the keys or a path'
              ' to a json file which which stores the needed information')
@click.option('--role', default='timestamp', help='Metadata role whose expiration date should be '
              'updated')
@click.option('--start-date', default=datetime.datetime.now(), help='Date to which the intercal is added', type=ISO_DATE)
@click.option('--interval', default=None, help='Time interval added to the start date', type=int)
@click.option('--commit-msg', default=None, help='Commit message to be used in case the changes'
              'should be automatically committed')
def update_expiration_date(repo_path, keystore, keys_description, role, start_date, interval,
                           commit_msg):
    developer_tool.update_metadata_expiration_date(repo_path, keystore, keys_description, role,
                                                   start_date, interval, commit_msg)


@cli.command()
@click.option('--url', help="Authentication repository's url")
@click.option('--clients-dir', help="Directory containing the client's authentication repository")
@click.option('--targets-dir', help="Directory containing the target repositories")
@click.option('--from-fs', is_flag=True, default=False, help='Indicates if the we want to clone a '
              'repository from the filesystem')
def update(url, clients_dir, targets_dir, from_fs):
    update_repository(url, clients_dir, targets_dir, from_fs)


@cli.command()
@click.option('--url', help="Authentication repository's url")
@click.option('--clients-dir', help="Directory containing the client's authentication repository")
@click.option('--repo-name', help="Repository's name")
@click.option('--targets-dir', help="Directory containing the target repositories")
@click.option('--from-fs', is_flag=True, default=False, help='Indicates if the we want to clone a '
              'repository from the filesystem')
def update_named_repo(url, clients_dir, repo_name, targets_dir, from_fs):
    update_named_repository(url, clients_dir, repo_name, targets_dir, from_fs)


@cli.command()
@click.option('--key-path', default=None, help='Path to the key which should be copied to a Yubikey')
def setup_test_yubikey(key_path):
    import taf.yubikey as yk

    if key_path is None:
        key_path = Path(__file__).parent.parent / "tests" / "data" / "keystore" / "targets"
    else:
        key_path = Path(key_path)

    key_pem = key_path.read_bytes()

    click.echo("\nImporting RSA private key from {} to Yubikey...".format(key_path))

    pin = yk.DEFAULT_PIN
    pub_key = yk.setup(pin, 'Test Yubikey', private_key_pem=key_pem)

    click.echo("\nPrivate key successfully imported.\n")
    click.echo("\nPublic key (PEM): \n{}".format(pub_key.decode("utf-8")))
    click.echo("Pin: {}\n".format(pin))


cli()



import os
import shutil
from contextlib import contextmanager
from pathlib import Path

import taf.repository_tool as repository_tool
import taf.yubikey
from pytest import fixture, yield_fixture
from securesystemslib.interface import (
    import_rsa_privatekey_from_file,
    import_rsa_publickey_from_file,
)
from taf.repository_tool import Repository
from taf.tests import TEST_WITH_REAL_YK
from taf.tests.yubikey_utils import (
    Root1YubiKey,
    Root2YubiKey,
    Root3YubiKey,
    TargetYubiKey,
    _yk_piv_ctrl_mock,
)
from taf.utils import on_rm_error

TEST_DATA_PATH = Path(__file__).parent / "data"
TEST_DATA_REPOS_PATH = TEST_DATA_PATH / "repos"
TEST_DATA_ORIGIN_PATH = TEST_DATA_REPOS_PATH / "origin"
KEYSTORE_PATH = TEST_DATA_PATH / "keystore"
WRONG_KEYSTORE_PATH = TEST_DATA_PATH / "wrong_keystore"
CLIENT_DIR_PATH = TEST_DATA_REPOS_PATH / "client"


def pytest_configure(config):
    if not TEST_WITH_REAL_YK:
        taf.yubikey._yk_piv_ctrl = _yk_piv_ctrl_mock


def pytest_generate_tests(metafunc):
    if "taf_happy_path" in metafunc.fixturenames:
        # When running tests with real yubikey, use just rsa-pkcs1v15-sha256 scheme
        schemes = (
            ["rsa-pkcs1v15-sha256"]
            if TEST_WITH_REAL_YK
            else ["rsassa-pss-sha256", "rsa-pkcs1v15-sha256"]
        )
        metafunc.parametrize("taf_happy_path", schemes, indirect=True)


@contextmanager
def origin_repos_group(test_group_dir):
    all_paths = {}
    test_group_dir = str(TEST_DATA_REPOS_PATH / test_group_dir)
    for test_dir in os.scandir(test_group_dir):
        if test_dir.is_dir():
            all_paths[test_dir.name] = _copy_repos(test_dir.path, test_dir.name)

    yield all_paths

    for test_name in all_paths:
        test_dst_path = str(TEST_DATA_ORIGIN_PATH / test_name)
        shutil.rmtree(test_dst_path, onerror=on_rm_error)


@contextmanager
def origin_repos(test_name):
    """Coppies git repository from `data/repos/test-XYZ` to data/repos/origin/test-XYZ
  path and renames `git` to `.git` for each repository.
  """
    test_dir_path = str(TEST_DATA_REPOS_PATH / test_name)
    temp_paths = _copy_repos(test_dir_path, test_name)

    yield temp_paths

    test_dst_path = str(TEST_DATA_ORIGIN_PATH / test_name)
    shutil.rmtree(test_dst_path, onerror=on_rm_error)


def _copy_repos(test_dir_path, test_name):
    paths = {}
    for root, dirs, _ in os.walk(test_dir_path):
        for dir_name in dirs:
            if dir_name == "git":
                repo_rel_path = Path(root).relative_to(test_dir_path)
                dst_path = TEST_DATA_ORIGIN_PATH / test_name / repo_rel_path
                # convert dst_path to string in order to support python 3.5
                shutil.copytree(root, str(dst_path))
                (dst_path / "git").rename(dst_path / ".git")
                repo_rel_path = Path(repo_rel_path).as_posix()
                paths[repo_rel_path] = str(dst_path)
    return paths


@yield_fixture(scope="session", autouse=True)
def taf_happy_path(request, pytestconfig):
    """TAF repository for testing."""
    repository_tool.DISABLE_KEYS_CACHING = True

    def _create_origin(test_dir, taf_repo_name="taf"):
        with origin_repos(test_dir) as origins:
            taf_repo_origin_path = origins[taf_repo_name]
            yield Repository(taf_repo_origin_path)

    scheme = request.param
    pytestconfig.option.signature_scheme = scheme

    if scheme == "rsassa-pss-sha256":
        yield from _create_origin("test-happy-path")
    elif scheme == "rsa-pkcs1v15-sha256":
        yield from _create_origin("test-happy-path-pkcs1v15")
    else:
        raise ValueError(f"Invalid test config. Invalid scheme: {scheme}")


@yield_fixture(scope="session", autouse=True)
def updater_repositories():
    test_dir = "test-updater"
    with origin_repos_group(test_dir) as origins:
        yield origins


@fixture
def client_dir():
    return CLIENT_DIR_PATH


@fixture
def origin_dir():
    return TEST_DATA_ORIGIN_PATH


@fixture
def keystore():
    """Keystore path."""
    return str(KEYSTORE_PATH)


@fixture
def wrong_keystore():
    """Path of the wrong keystore"""
    return str(WRONG_KEYSTORE_PATH)


@fixture
def targets_yk(pytestconfig):
    """Targets YubiKey."""
    return TargetYubiKey(KEYSTORE_PATH, pytestconfig.option.signature_scheme)


@fixture
def root1_yk(pytestconfig):
    """Root1 YubiKey."""
    return Root1YubiKey(KEYSTORE_PATH, pytestconfig.option.signature_scheme)


@fixture
def root2_yk(pytestconfig):
    """Root2 YubiKey."""
    return Root2YubiKey(KEYSTORE_PATH, pytestconfig.option.signature_scheme)


@fixture
def root3_yk(pytestconfig):
    """Root3 YubiKey."""
    return Root3YubiKey(KEYSTORE_PATH, pytestconfig.option.signature_scheme)


@fixture
def snapshot_key(pytestconfig):
    """Snapshot key."""
    key = import_rsa_publickey_from_file(
        str(KEYSTORE_PATH / "snapshot.pub"), scheme=pytestconfig.option.signature_scheme
    )
    priv_key = import_rsa_privatekey_from_file(
        str(KEYSTORE_PATH / "snapshot"), scheme=pytestconfig.option.signature_scheme
    )
    key["keyval"]["private"] = priv_key["keyval"]["private"]
    return key


@fixture
def timestamp_key(pytestconfig):
    """Timestamp key."""
    key = import_rsa_publickey_from_file(
        str(KEYSTORE_PATH / "timestamp.pub"),
        scheme=pytestconfig.option.signature_scheme,
    )
    priv_key = import_rsa_privatekey_from_file(
        str(KEYSTORE_PATH / "timestamp"), scheme=pytestconfig.option.signature_scheme
    )
    key["keyval"]["private"] = priv_key["keyval"]["private"]
    return key


@yield_fixture
def targets_key(pytestconfig):
    """Targets key."""
    key = import_rsa_publickey_from_file(
        str(KEYSTORE_PATH / "targets.pub"), scheme=pytestconfig.option.signature_scheme
    )
    priv_key = import_rsa_privatekey_from_file(
        str(KEYSTORE_PATH / "targets"), scheme=pytestconfig.option.signature_scheme
    )
    key["keyval"]["private"] = priv_key["keyval"]["private"]
    return key



import datetime
import random
from contextlib import contextmanager

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from securesystemslib.pyca_crypto_keys import create_rsa_signature
from tuf.repository_tool import import_rsakey_from_pem
from ykman.descriptor import FailedOpeningDeviceException
from ykman.piv import WrongPin

VALID_PIN = "123456"
WRONG_PIN = "111111"

INSERTED_YUBIKEY = None


class FakeYubiKey:
    def __init__(self, priv_key_path, pub_key_path, scheme, serial=None, pin=VALID_PIN):
        self.priv_key_pem = priv_key_path.read_bytes()
        self.pub_key_pem = pub_key_path.read_bytes()

        self._serial = serial if serial else random.randint(100000, 999999)
        self._pin = pin

        self.scheme = scheme
        self.priv_key = serialization.load_pem_private_key(
            self.priv_key_pem, None, default_backend()
        )
        self.pub_key = serialization.load_pem_public_key(
            self.pub_key_pem, default_backend()
        )

        self.tuf_key = import_rsakey_from_pem(self.pub_key_pem.decode("utf-8"), scheme)

    @property
    def driver(self):
        return self

    @property
    def pin(self):
        return self._pin

    @property
    def serial(self):
        return self._serial

    def insert(self):
        """Insert YubiKey in USB slot."""
        global INSERTED_YUBIKEY
        INSERTED_YUBIKEY = self

    def is_inserted(self):
        """Check if YubiKey is in USB slot."""
        global INSERTED_YUBIKEY
        return INSERTED_YUBIKEY is self

    def remove(self):
        """Removes YubiKey from USB slot."""
        global INSERTED_YUBIKEY
        if INSERTED_YUBIKEY is self:
            INSERTED_YUBIKEY = None


class FakePivController:
    def __init__(self, driver):
        self._driver = driver

    @property
    def driver(self):
        return None

    def authenticate(self, *args, **kwargs):
        pass

    def change_pin(self, *args, **kwargs):
        pass

    def change_puk(self, *args, **kwargs):
        pass

    def generate_self_signed_certificate(self, *args, **kwargs):
        pass

    def read_certificate(self, _slot):
        name = x509.Name(
            [x509.NameAttribute(x509.NameOID.COMMON_NAME, self.__class__.__name__)]
        )
        now = datetime.datetime.utcnow()

        return (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(self._driver.pub_key)
            .serial_number(self._driver.serial)
            .not_valid_before(now)
            .not_valid_after(now + datetime.timedelta(days=365))
            .sign(self._driver.priv_key, hashes.SHA256(), default_backend())
        )

    def reset(self):
        pass

    def set_pin_retries(self, *args, **kwargs):
        pass

    def sign(self, slot, algorithm, data):
        """Sign data using the same function as TUF"""
        if isinstance(data, str):
            data = data.encode("utf-8")

        sig, _ = create_rsa_signature(
            self._driver.priv_key_pem.decode("utf-8"), data, self._driver.scheme
        )
        return sig

    def verify(self, pin):
        if self._driver.pin != pin:
            raise WrongPin("", "")


class TargetYubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(
            keystore_path / "targets", keystore_path / "targets.pub", scheme
        )


class Root1YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root1", keystore_path / "root1.pub", scheme)


class Root2YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root2", keystore_path / "root2.pub", scheme)


class Root3YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root3", keystore_path / "root3.pub", scheme)


@contextmanager
def _yk_piv_ctrl_mock(serial=None, pub_key_pem=None):
    global INSERTED_YUBIKEY

    if INSERTED_YUBIKEY is None:
        raise FailedOpeningDeviceException()

    yield FakePivController(INSERTED_YUBIKEY), INSERTED_YUBIKEY.serial



import datetime
from contextlib import contextmanager
from functools import wraps

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from tuf.repository_tool import import_rsakey_from_pem
from ykman.descriptor import list_devices, open_device
from ykman.piv import (
    ALGO,
    DEFAULT_MANAGEMENT_KEY,
    PIN_POLICY,
    SLOT,
    PivController,
    WrongPin,
    generate_random_management_key,
)
from ykman.util import TRANSPORT

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import YubikeyError

DEFAULT_PIN = "123456"
DEFAULT_PUK = "12345678"


def raise_yubikey_err(msg=None):
    """Decorator used to catch all errors raised by yubikey-manager and raise
    YubikeyError. We don't need to handle specific cases.
    """

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except YubikeyError:
                raise
            except Exception as e:
                err_msg = (
                    f"{msg} Reason: ({type(e).__name__}) {str(e)}" if msg else str(e)
                )
                raise YubikeyError(err_msg) from e

        return decorator

    return wrapper


@contextmanager
def _yk_piv_ctrl(serial=None, pub_key_pem=None):
    """Context manager to open connection and instantiate piv controller.

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        - ykman.piv.PivController

    Raises:
        - YubikeyError
    """
    # If pub_key_pem is given, iterate all devices, read x509 certs and try to match
    # public keys.
    if pub_key_pem is not None:
        for yk in list_devices(transports=TRANSPORT.CCID):
            yk_ctrl = PivController(yk.driver)
            device_pub_key_pem = (
                yk_ctrl.read_certificate(SLOT.SIGNATURE)
                .public_key()
                .public_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                )
                .decode("utf-8")
            )
            # Tries to match without last newline char
            if (
                device_pub_key_pem == pub_key_pem
                or device_pub_key_pem[:-1] == pub_key_pem
            ):
                break
            else:
                yk.close()

    else:
        yk = open_device(transports=TRANSPORT.CCID, serial=serial)
        yk_ctrl = PivController(yk.driver)

    yield yk_ctrl, yk.serial
    yk.close()


def is_inserted():
    """Checks if YubiKey is inserted.

    Args:
        None

    Returns:
        True if at least one Yubikey is inserted (bool)

    Raises:
        - YubikeyError
    """
    return len(list(list_devices(transports=TRANSPORT.CCID))) > 0


@raise_yubikey_err()
def is_valid_pin(pin):
    """Checks if given pin is valid.

    Args:
        pin(str): Yubikey piv PIN

    Returns:
        tuple: True if PIN is valid, otherwise False, number of PIN retries

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        try:
            ctrl.verify(pin)
            return True, None  # ctrl.get_pin_tries() fails if PIN is valid
        except WrongPin:
            return False, ctrl.get_pin_tries()


@raise_yubikey_err("Cannot get serial number.")
def get_serial_num(pub_key_pem=None):
    """Get Yubikey serial number.

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Yubikey serial number

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (_, serial):
        return serial


@raise_yubikey_err("Cannot export x509 certificate.")
def export_piv_x509(cert_format=serialization.Encoding.PEM, pub_key_pem=None):
    """Exports YubiKey's piv slot x509.

    Args:
        - cert_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV x509 certificate in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        x509 = ctrl.read_certificate(SLOT.SIGNATURE)
        return x509.public_bytes(encoding=cert_format)


@raise_yubikey_err("Cannot export public key.")
def export_piv_pub_key(pub_key_format=serialization.Encoding.PEM, pub_key_pem=None):
    """Exports YubiKey's piv slot public key.

    Args:
        - pub_key_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV public key in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        x509 = ctrl.read_certificate(SLOT.SIGNATURE)
        return x509.public_key().public_bytes(
            encoding=pub_key_format,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )


@raise_yubikey_err("Cannot get public key in TUF format.")
def get_piv_public_key_tuf(scheme=DEFAULT_RSA_SIGNATURE_SCHEME, pub_key_pem=None):
    """Return public key from a Yubikey in TUF's RSAKEY_SCHEMA format.

    Args:
        - scheme(str): Rsa signature scheme (default is rsa-pkcs1v15-sha256)
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        A dictionary containing the RSA keys and other identifying information
        from inserted smart card.
        Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.

    Raises:
        - YubikeyError
    """
    pub_key_pem = export_piv_pub_key(pub_key_pem=pub_key_pem).decode("utf-8")
    return import_rsakey_from_pem(pub_key_pem, scheme)


@raise_yubikey_err("Cannot sign data.")
def sign_piv_rsa_pkcs1v15(data, pin, pub_key_pem=None):
    """Sign data with key from YubiKey's piv slot.

    Args:
        - data(bytes): Data to be signed
        - pin(str): Pin for piv slot login.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Signature (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        ctrl.verify(pin)
        return ctrl.sign(SLOT.SIGNATURE, ALGO.RSA2048, data)


@raise_yubikey_err("Cannot setup Yubikey.")
def setup(
    pin,
    cert_cn,
    cert_exp_days=365,
    pin_retries=10,
    private_key_pem=None,
    mgm_key=generate_random_management_key(),
):
    """Use to setup inserted Yubikey, with following steps (order is important):
      - reset to factory settings
      - set management key
      - generate key(RSA2048) or import given one
      - generate and import self-signed certificate(X509)
      - set pin retries
      - set pin
      - set puk(same as pin)

    Args:
        - cert_cn(str): x509 common name
        - cert_exp_days(int): x509 expiration (in days from now)
        - pin_retries(int): Number of retries for PIN
        - private_key_pem(str): Private key in PEM format. If given, it will be
                                imported to Yubikey.
        - mgm_key(bytes): New management key

    Returns:
        PIV public key in PEM format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        # Factory reset and set PINs
        ctrl.reset()

        ctrl.authenticate(DEFAULT_MANAGEMENT_KEY)
        ctrl.set_mgm_key(mgm_key)

        # Generate RSA2048
        if private_key_pem is None:
            pub_key = ctrl.generate_key(SLOT.SIGNATURE, ALGO.RSA2048, PIN_POLICY.ALWAYS)
        else:
            private_key = load_pem_private_key(private_key_pem, None, default_backend())
            ctrl.import_key(SLOT.SIGNATURE, private_key, PIN_POLICY.ALWAYS)
            pub_key = private_key.public_key()

        ctrl.authenticate(mgm_key)
        ctrl.verify(DEFAULT_PIN)

        # Generate and import certificate
        now = datetime.datetime.now()
        valid_to = now + datetime.timedelta(days=cert_exp_days)
        ctrl.generate_self_signed_certificate(
            SLOT.SIGNATURE, pub_key, cert_cn, now, valid_to
        )

        ctrl.set_pin_retries(pin_retries=pin_retries, puk_retries=pin_retries)
        ctrl.change_pin(DEFAULT_PIN, pin)
        ctrl.change_puk(DEFAULT_PUK, pin)

    return pub_key.public_bytes(
        serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
    )



from setuptools import find_packages, setup

PACKAGE_NAME = "taf"
VERSION = "0.1.8"
AUTHOR = "Open Law Library"
AUTHOR_EMAIL = "info@openlawlib.org"
DESCRIPTION = "Implementation of archival authentication"
KEYWORDS = "update updater secure authentication archival"
URL = "https://github.com/openlawlibrary/taf/tree/master"

with open("README.md", encoding="utf-8") as file_object:
    long_description = file_object.read()

packages = find_packages()

# Create platform specific wheel
# https://stackoverflow.com/a/45150383/9669050
try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False


except ImportError:
    bdist_wheel = None

ci_require = [
    "bandit>=1.6.0",
    "black>=19.3b0",
    "coverage==4.5.3",
    "pre-commit>=1.18.3",
    "pytest-cov==2.7.1",
]

dev_require = ["bandit>=1.6.0", "black>=19.3b0", "pre-commit>=1.18.3"]

tests_require = ["pytest==4.5.0"]

yubikey_require = ["yubikey-manager==3.0.0"]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    keywords=KEYWORDS,
    packages=packages,
    cmdclass={"bdist_wheel": bdist_wheel},
    include_package_data=True,
    data_files=[("lib/site-packages/taf", ["./LICENSE.txt", "./README.md"])],
    zip_safe=False,
    install_requires=[
        "click==6.7",
        "colorama>=0.3.9",
        "cryptography>=2.3.1",
        "oll-tuf==0.11.2.dev9",
    ],
    extras_require={
        "ci": ci_require,
        "test": tests_require,
        "dev": dev_require,
        "yubikey": yubikey_require,
    },
    tests_require=tests_require,
    entry_points={
        "console_scripts": ["taf = taf.cli:main"],
        "pytest11": ["taf_yubikey_utils = taf.tests.yubikey_utils"],
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Security",
        "Topic :: Software Development",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



"""
This module includes tests for the taf updater. The basic idea behind the tests is the following:
1. For each situation we want to test, there is a set of target repositories and an authentication repository.
For example, repositories where everything is completely valid, repositories where there is a mismatch between
a target repository's commit and the corresponding target commit sha in the auth repo etc.
These repositories are placed in tests/data/repos/test-updater directory and are copied to the origin directory
before any of the tests are executed. The repositories in the origin directory are used as remotes.
As a future improvements, we could reduce the number of repositories by adding/removing commits from tests.
2. We test update when no client repositories exist and when they do exist, but are not up to date.
When testing the second case, we clone the repositories and then revert them. We only modify client's repositories.
3. When everything is set up, update is called. Afterwards:
  1) If the update is expected to be successful, it is checked if pulled commits match commits of the remote
   repositories and if last_successful commit was created.
  2) If the update is expected to be unsuccessful, it is checked if the client's repositories remained the same
  (if they existed) or if they do not exist (in case they didn't exist in the first place) and if the
  error message is as expected.
4. Client repositories are deleted after every test to make sure that the execution of a previous test does
not impact the following ones.

These states of repositories are tested:
1. The authentication and target repositories are valid and all of the commits match - test-updater-valid. This
set of repositories is also used to test if the updater works in case when there is no need to perform the update,
and if the update fails if there are no client target repositories, but the auth repo exists.
2. There are additional commits in target repos (commits which are not accounted for by the auth repo) -
test-updater-additional-target-commit The update is performed until the last commit that can be validated.
Updated does not raise an error (a warning is logged).
3. The authentication and target repositories are valid and all of the commits match, but there are also auth
repo commits where only metadata files were updated (e.g. timestamp expiration date was changed) -
test-updater-valid-with-updated-expiration-dates. The update should be successful.
4. There are more commits noted in auth repo's target files than target repositories' commits
test-updater-missing-target-commit. The update is expected to fail.
5. The commits of the target repositories do not match commits noted in the auth repo - test-updater-invalid-target-sha.
The update is expected to fail.
6. A metadata file's signature is not valid - test-updater-wrong-key. The update is expected to fail.
7. A metadata file's expiration date is not valid - test-updater-invalid-expiration-date.
The update is expected to fail.
8. A metadata file's version number is not valid (this test case was created by swapping commits using
git rebase) - test-updater-invalid-version-number. The update is expected to fail.
9. A metadata file which should have remained the same changed (this test case was created by updating targets
and not updating snaphost) - test-updater-just-targets-updated. The update is expected to fail.

On top of that, it is tested that update fails if last_validated_commit does not exist, while the client's repository
does and if it does exist, but the stored commit does not match the client repository's head commit.
"""

import os
import shutil
from pathlib import Path

import pytest
from pytest import fixture

import taf.settings as settings
from taf.auth_repo import AuthenticationRepo
from taf.exceptions import UpdateFailedError
from taf.git import GitRepository
from taf.updater.updater import update_repository
from taf.utils import on_rm_error

AUTH_REPO_REL_PATH = "organization/auth_repo"
TARGET1_SHA_MISMATCH = "Mismatch between target commits specified in authentication repository and target repository namespace/TargetRepo1"
TARGETS_MISMATCH_ANY = "Mismatch between target commits specified in authentication repository and target repository"
NO_WORKING_MIRRORS = "Validation of authentication repository auth_repo failed due to error: No working mirror was found"
TIMESTAMP_EXPIRED = "Metadata 'timestamp' expired"
REPLAYED_METADATA = "ReplayedMetadataError"
METADATA_CHANGED_BUT_SHOULDNT = (
    "Metadata file targets.json should be the same at revisions"
)


def setup_module(module):
    settings.update_from_filesystem = True


def teardown_module(module):
    settings.update_from_filesystem = False


@fixture(autouse=True)
def run_around_tests(client_dir):
    yield
    for root, dirs, _ in os.walk(str(client_dir)):
        for dir_name in dirs:
            shutil.rmtree(str(Path(root) / dir_name), onerror=on_rm_error)


@pytest.mark.parametrize(
    "test_name, test_repo",
    [
        ("test-updater-valid", False),
        ("test-updater-additional-target-commit", False),
        ("test-updater-valid-with-updated-expiration-dates", False),
        ("test-updater-allow-unauthenticated-commits", False),
        ("test-updater-test-repo", True),
    ],
)
def test_valid_update_no_client_repo(
    test_name, test_repo, updater_repositories, origin_dir, client_dir
):
    repositories = updater_repositories[test_name]
    origin_dir = origin_dir / test_name
    _update_and_check_commit_shas(None, repositories, origin_dir, client_dir, test_repo)


@pytest.mark.parametrize(
    "test_name, num_of_commits_to_revert",
    [
        ("test-updater-valid", 3),
        ("test-updater-additional-target-commit", 1),
        ("test-updater-allow-unauthenticated-commits", 1),
    ],
)
def test_valid_update_existing_client_repos(
    test_name, num_of_commits_to_revert, updater_repositories, origin_dir, client_dir
):
    # clone the origin repositories
    # revert them to an older commit
    repositories = updater_repositories[test_name]
    origin_dir = origin_dir / test_name
    client_repos = _clone_and_revert_client_repositories(
        repositories, origin_dir, client_dir, num_of_commits_to_revert
    )
    # create valid last validated commit file
    _create_last_validated_commit(
        client_dir, client_repos[AUTH_REPO_REL_PATH].head_commit_sha()
    )
    _update_and_check_commit_shas(client_repos, repositories, origin_dir, client_dir)


@pytest.mark.parametrize(
    "test_name, test_repo",
    [
        ("test-updater-valid", False),
        ("test-updater-allow-unauthenticated-commits", False),
        ("test-updater-test-repo", True),
    ],
)
def test_no_update_necessary(
    test_name, test_repo, updater_repositories, origin_dir, client_dir
):
    # clone the origin repositories
    # revert them to an older commit
    repositories = updater_repositories[test_name]
    origin_dir = origin_dir / test_name
    client_repos = _clone_client_repositories(repositories, origin_dir, client_dir)
    # create valid last validated commit file
    _create_last_validated_commit(
        client_dir, client_repos[AUTH_REPO_REL_PATH].head_commit_sha()
    )
    _update_and_check_commit_shas(
        client_repos, repositories, origin_dir, client_dir, test_repo
    )


@pytest.mark.parametrize(
    "test_name, expected_error",
    [
        ("test-updater-invalid-target-sha", TARGET1_SHA_MISMATCH),
        ("test-updater-missing-target-commit", TARGET1_SHA_MISMATCH),
        ("test-updater-wrong-key", NO_WORKING_MIRRORS),
        ("test-updater-invalid-expiration-date", TIMESTAMP_EXPIRED),
        ("test-updater-invalid-version-number", REPLAYED_METADATA),
        ("test-updater-just-targets-updated", METADATA_CHANGED_BUT_SHOULDNT),
    ],
)
def test_updater_invalid_update(
    test_name, expected_error, updater_repositories, client_dir
):
    repositories = updater_repositories[test_name]
    clients_auth_repo_path = client_dir / AUTH_REPO_REL_PATH
    _update_invalid_repos_and_check_if_repos_exist(
        client_dir, repositories, expected_error
    )
    # make sure that the last validated commit does not exist
    _check_if_last_validated_commit_exists(clients_auth_repo_path)


@pytest.mark.parametrize(
    "test_name, expected_error",
    [("test-updater-invalid-target-sha", TARGET1_SHA_MISMATCH)],
)
def test_updater_invalid_target_sha_existing_client_repos(
    test_name, expected_error, updater_repositories, origin_dir, client_dir
):
    repositories = updater_repositories[test_name]
    clients_auth_repo_path = client_dir / AUTH_REPO_REL_PATH
    origin_dir = origin_dir / test_name
    client_repos = _clone_and_revert_client_repositories(
        repositories, origin_dir, client_dir, 1
    )
    _create_last_validated_commit(
        client_dir, client_repos[AUTH_REPO_REL_PATH].head_commit_sha()
    )
    _update_invalid_repos_and_check_if_remained_same(
        client_repos, client_dir, repositories, expected_error
    )
    _check_last_validated_commit(clients_auth_repo_path)


def test_no_target_repositories(updater_repositories, origin_dir, client_dir):
    repositories = updater_repositories["test-updater-valid"]
    origin_dir = origin_dir / "test-updater-valid"
    client_auth_repo = _clone_client_repo(AUTH_REPO_REL_PATH, origin_dir, client_dir)
    _create_last_validated_commit(client_dir, client_auth_repo.head_commit_sha())
    client_repos = {AUTH_REPO_REL_PATH: client_auth_repo}
    _update_invalid_repos_and_check_if_remained_same(
        client_repos, client_dir, repositories, TARGETS_MISMATCH_ANY
    )
    # make sure that the target repositories still do not exist
    for repository_rel_path in repositories:
        if repository_rel_path != AUTH_REPO_REL_PATH:
            client_repo_path = client_dir / repository_rel_path
            assert client_repo_path.exists() is False


def test_no_last_validated_commit(updater_repositories, origin_dir, client_dir):
    # clone the origin repositories
    # revert them to an older commit
    repositories = updater_repositories["test-updater-valid"]
    origin_dir = origin_dir / "test-updater-valid"
    client_repos = _clone_and_revert_client_repositories(
        repositories, origin_dir, client_dir, 3
    )
    # update without setting the last validated commit
    # update should start from the beginning and be successful
    _update_and_check_commit_shas(client_repos, repositories, origin_dir, client_dir)


def test_invalid_last_validated_commit(updater_repositories, origin_dir, client_dir):
    # clone the origin repositories
    # revert them to an older commit
    repositories = updater_repositories["test-updater-valid"]
    origin_dir = origin_dir / "test-updater-valid"
    client_repos = _clone_and_revert_client_repositories(
        repositories, origin_dir, client_dir, 3
    )
    first_commit = client_repos[AUTH_REPO_REL_PATH].all_commits_on_branch()[0]
    expected_error = (
        f"Saved last validated commit {first_commit} does not match the head commit"
    )
    _create_last_validated_commit(client_dir, first_commit)
    # try to update without setting the last validated commit
    _update_invalid_repos_and_check_if_remained_same(
        client_repos, client_dir, repositories, expected_error
    )


def test_update_test_repo_no_flag(updater_repositories, origin_dir, client_dir):
    repositories = updater_repositories["test-updater-test-repo"]
    origin_dir = origin_dir / "test-updater-test-repo"
    expected_error = "Repository auth_repo is a test repository."
    # try to update without setting the last validated commit
    _update_invalid_repos_and_check_if_repos_exist(
        client_dir, repositories, expected_error
    )


def test_update_repo_wrong_flag(updater_repositories, origin_dir, client_dir):
    repositories = updater_repositories["test-updater-valid"]
    origin_dir = origin_dir / "test-updater-valid"
    expected_error = "Repository auth_repo is not a test repository."
    # try to update without setting the last validated commit
    _update_invalid_repos_and_check_if_repos_exist(
        client_dir, repositories, expected_error, True
    )


def _check_last_validated_commit(clients_auth_repo_path):
    # check if last validated commit is created and the saved commit is correct
    client_auth_repo = AuthenticationRepo(
        str(clients_auth_repo_path), "metadata", "targets"
    )
    head_sha = client_auth_repo.head_commit_sha()
    last_validated_commit = client_auth_repo.last_validated_commit
    assert head_sha == last_validated_commit


def _check_if_last_validated_commit_exists(clients_auth_repo_path):
    client_auth_repo = AuthenticationRepo(
        str(clients_auth_repo_path), "metadata", "targets"
    )
    last_validated_commit = client_auth_repo.last_validated_commit
    assert last_validated_commit is None


def _check_if_commits_match(repositories, origin_dir, client_dir, start_head_shas=None):
    for repository_rel_path in repositories:
        origin_repo = GitRepository(origin_dir / repository_rel_path)
        client_repo = GitRepository(client_dir / repository_rel_path)
        if start_head_shas is not None:
            start_commit = start_head_shas.get(repository_rel_path)
        else:
            start_commit = None
        origin_auth_repo_commits = origin_repo.all_commits_since_commit(start_commit)
        client_auth_repo_commits = client_repo.all_commits_since_commit(start_commit)
        for origin_commit, client_commit in zip(
            origin_auth_repo_commits, client_auth_repo_commits
        ):
            assert origin_commit == client_commit


def _clone_client_repositories(repositories, origin_dir, client_dir):
    client_repos = {}
    for repository_rel_path in repositories:
        client_repo = _clone_client_repo(repository_rel_path, origin_dir, client_dir)
        client_repos[repository_rel_path] = client_repo
    return client_repos


def _clone_and_revert_client_repositories(
    repositories, origin_dir, client_dir, num_of_commits
):
    client_repos = {}

    client_auth_repo = _clone_client_repo(AUTH_REPO_REL_PATH, origin_dir, client_dir)
    client_auth_repo.reset_num_of_commits(num_of_commits, True)
    client_auth_repo_head_sha = client_auth_repo.head_commit_sha()
    client_repos[AUTH_REPO_REL_PATH] = client_auth_repo

    for repository_rel_path in repositories:
        if repository_rel_path == AUTH_REPO_REL_PATH:
            continue

        client_repo = _clone_client_repo(repository_rel_path, origin_dir, client_dir)
        # read the commit sha stored in target files
        commit = client_auth_repo.get_json(
            client_auth_repo_head_sha,
            str((Path("targets") / repository_rel_path).as_posix()),
        )
        commit_sha = commit["commit"]
        client_repo.reset_to_commit(commit_sha, True)
        client_repos[repository_rel_path] = client_repo

    return client_repos


def _clone_client_repo(repository_rel_path, origin_dir, client_dir):
    origin_repo_path = str(origin_dir / repository_rel_path)
    client_repo_path = str(client_dir / repository_rel_path)
    client_repo = GitRepository(client_repo_path, [origin_repo_path])
    client_repo.clone()
    return client_repo


def _create_last_validated_commit(client_dir, client_auth_repo_head_sha):
    client_conf_repo = client_dir / "organization/_auth_repo"
    client_conf_repo.mkdir(parents=True, exist_ok=True)
    with open(str(client_conf_repo / "last_validated_commit"), "w") as f:
        f.write(client_auth_repo_head_sha)


def _update_and_check_commit_shas(
    client_repos, repositories, origin_dir, client_dir, authetnicate_test_repo=False
):
    if client_repos is not None:
        start_head_shas = {
            repo_rel_path: repo.head_commit_sha()
            for repo_rel_path, repo in client_repos.items()
        }
    else:
        start_head_shas = {repo_rel_path: None for repo_rel_path in repositories}

    clients_auth_repo_path = client_dir / AUTH_REPO_REL_PATH
    origin_auth_repo_path = repositories[AUTH_REPO_REL_PATH]
    update_repository(
        str(origin_auth_repo_path),
        str(clients_auth_repo_path),
        str(client_dir),
        True,
        authenticate_test_repo=authetnicate_test_repo,
    )
    _check_if_commits_match(repositories, origin_dir, client_dir, start_head_shas)
    _check_last_validated_commit(clients_auth_repo_path)


def _update_invalid_repos_and_check_if_remained_same(
    client_repos, client_dir, repositories, expected_error, authenticate_test_repo=False
):

    start_head_shas = {
        repo_rel_path: repo.head_commit_sha()
        for repo_rel_path, repo in client_repos.items()
    }
    clients_auth_repo_path = client_dir / AUTH_REPO_REL_PATH
    origin_auth_repo_path = repositories[AUTH_REPO_REL_PATH]

    with pytest.raises(UpdateFailedError, match=expected_error):
        update_repository(
            str(origin_auth_repo_path),
            str(clients_auth_repo_path),
            str(client_dir),
            True,
            authenticate_test_repo=authenticate_test_repo,
        )

    # all repositories should still have the same head commit
    for repo_path, repo in client_repos.items():
        current_head = repo.head_commit_sha()
        assert current_head == start_head_shas[repo_path]


def _update_invalid_repos_and_check_if_repos_exist(
    client_dir, repositories, expected_error, authenticate_test_repo=False
):

    clients_auth_repo_path = client_dir / AUTH_REPO_REL_PATH
    origin_auth_repo_path = repositories[AUTH_REPO_REL_PATH]
    with pytest.raises(UpdateFailedError, match=expected_error):
        update_repository(
            str(origin_auth_repo_path),
            str(clients_auth_repo_path),
            str(client_dir),
            True,
            authenticate_test_repo=authenticate_test_repo,
        ),

    # the client repositories should not exits
    for repository_rel_path in repositories:
        path = client_dir / repository_rel_path
        assert path.exists() is False
